import random

import numpy as np
import torch
import torch.nn as nn
from forecasting_models.autoformer import AutoCorrelation


class MultiHeadAttention(nn.Module):

    def __init__(self, d_model, d_k, d_v, n_heads, attn_type, seed):

        super(MultiHeadAttention, self).__init__()

        np.random.seed(seed)
        random.seed(seed)
        torch.manual_seed(seed)

        self.WQ = nn.Linear(d_model, d_k * n_heads, bias=False)
        self.WK = nn.Linear(d_model, d_k * n_heads, bias=False)
        self.WV = nn.Linear(d_model, d_v * n_heads, bias=False)
        self.fc = nn.Linear(n_heads * d_v, d_model, bias=False)

        self.d_model = d_model
        self.d_k = d_k
        self.d_v = d_v
        self.n_heads = n_heads
        self.attn_type = attn_type
        self.seed = seed

    def forward(self, Q, K, V):

        batch_size = Q.shape[0]
        q_s = self.WQ(Q).view(batch_size, -1, self.n_heads, self.d_k).transpose(1, 2)
        k_s = self.WK(K).view(batch_size, -1, self.n_heads, self.d_k).transpose(1, 2)
        v_s = self.WV(V).view(batch_size, -1, self.n_heads, self.d_v).transpose(1, 2)

        context, attn = AutoCorrelation(seed=self.seed)(q_s.transpose(1, 2),
                                                        k_s.transpose(1, 2),
                                                        v_s.transpose(1, 2))

        context = context.transpose(1, 2).contiguous().view(batch_size, -1, self.n_heads * self.d_v)
        outputs = self.fc(context)
        return outputs


import random

import numpy as np
import torch
import torch.nn as nn
from modules.multi_head_attention import MultiHeadAttention
from modules.feedforward import PoswiseFeedForwardNet
from modules.encoding import PositionalEncoding


class EncoderLayer(nn.Module):

    def __init__(self, d_model, d_ff, d_k, d_v, n_heads,
                 attn_type, seed):
        super(EncoderLayer, self).__init__()

        np.random.seed(seed)
        random.seed(seed)
        torch.manual_seed(seed)

        self.enc_self_attn = MultiHeadAttention(
            d_model=d_model, d_k=d_k,
            d_v=d_v, n_heads=n_heads,
            attn_type=attn_type, seed=seed)

        self.pos_ffn = PoswiseFeedForwardNet(
            d_model=d_model, d_ff=d_ff)
        self.layer_norm = nn.LayerNorm(d_model, elementwise_affine=False)

    def forward(self, enc_inputs):

        out = self.enc_self_attn(enc_inputs, enc_inputs, enc_inputs)
        out = self.layer_norm(out + enc_inputs)
        out_2 = self.pos_ffn(out)
        out_2 = self.layer_norm(out_2 + out)
        return out_2


class Encoder(nn.Module):

    def __init__(self, d_model, d_ff, d_k, d_v, n_heads,
                 n_layers, device,
                 attn_type, seed):
        super(Encoder, self).__init__()

        np.random.seed(seed)
        random.seed(seed)
        torch.manual_seed(seed)

        self.device = device
        self.attn_type = attn_type
        self.pos_emb = PositionalEncoding(
            d_hid=d_model, device=device)
        self.n_layers = n_layers
        self.layers = []
        for _ in range(n_layers):
            encoder_layer = EncoderLayer(
                d_model=d_model, d_ff=d_ff,
                d_k=d_k, d_v=d_v, n_heads=n_heads,
                attn_type=attn_type, seed=seed)
            self.layers.append(encoder_layer)
        self.layers = nn.ModuleList(self.layers)

    def forward(self, enc_input):

        enc_outputs = self.pos_emb(enc_input)

        for layer in self.layers:
            enc_outputs = layer(enc_outputs)

        return enc_outputs


import torch.nn as nn
import torch


class PositionalEncoding(nn.Module):
    """Positional encoding."""
    def __init__(self, d_hid, device, max_len=1000):

        super(PositionalEncoding, self).__init__()
        # Create a long enough `P`
        self.P = torch.zeros((1, max_len, d_hid), device=device)
        X = torch.arange(max_len, dtype=torch.float32).reshape(
            -1, 1) / torch.pow(10000, torch.arange(
            0, d_hid, 2, dtype=torch.float32) / d_hid)
        self.P[:, :, 0::2] = torch.sin(X).to(device)
        self.P[:, :, 1::2] = torch.cos(X).to(device)

    def forward(self, X):
        X = X + self.P[:, :X.shape[1], :]
        return X



import os
import random
import pandas as pd
import numpy as np
import torch
import torch.nn as nn
import optuna
import torch.nn.functional as F
from optuna.trial import TrialState
from torch.optim import Adam
from forecast_blur_denoise import ForecastBlurDenoise


class TrainForecastBlurDenoise:
    def __init__(self,
                 *,
                 exp_name='toy_data',
                 forecasting_model_name="LSTM",
                 n_jobs=1,
                 n_trials=5,
                 num_epochs=10,
                 num_iteration,
                 forecasting_model,
                 train,
                 valid,
                 test,
                 noise_type,
                 add_noise_only_at_training,
                 input_size,
                 output_size,
                 pred_len,
                 num_inducing,
                 hyperparameters,
                 seed,
                 device):
        """
        Trainer class for training the ForecastDenoising model using Optuna hyperparameter optimization.

        Args:
        - exp_name (str): Name of the experiment (dataset).
        - forecating_model_name (str): Name of the forecasting model.
        - n_jobs (int): Total number of jobs for Optuna.
        - num_epochs (int): Total number of epochs.
        - forecasting_model (nn.Module): The underlying forecasting model.
        - train (DataLoader): DataLoader for training data.
        - valid (DataLoader): DataLoader for validation data.
        - test (DataLoader): DataLoader for test data.
        - noise_type (str): Type of noise to be added during denoising ('gp', 'iso', 'no_noise').
        - add_noise_only_at_training (bool): Flag indicating whether to add noise only during training.
        - src_input_size (int): Size of the source input.
        - tgt_input_size (int): Size of the target input.
        - tgt_output_size (int): Size of the target output.
        - pred_len (int): Length of the prediction horizon.
        - num_inducing (int): Number of inducing points for GP regression.
        - hyperparameters (dict): Hyperparameters to be optimized.
        - args: Command line arguments.
        - seed (int): Random seed for reproducibility.
        - device: Device on which to run the training.
        """

        random.seed(seed)
        np.random.seed(seed)
        torch.random.manual_seed(seed)

        self.n_jobs = n_jobs
        self.n_trials = n_trials
        self.num_epochs = num_epochs
        self.exp_name = exp_name
        self.forecasting_model_name = forecasting_model_name

        gp = True if noise_type == "gp" else False
        iso = True if noise_type == "iso" else False
        no_noise = True if noise_type == "no_noise" else False

        self.train_data, self.valid_data, self.test_data = train, valid, test
        self.forecasting_model = forecasting_model

        self.forecast_denoising_model = ForecastBlurDenoise(forecasting_model=forecasting_model,
                                                            gp=gp, iso=iso, no_noise=no_noise,
                                                            add_noise_only_at_training=add_noise_only_at_training,
                                                            input_size=input_size,
                                                            output_size=output_size,
                                                            pred_len=pred_len,
                                                            num_inducing=num_inducing).to(device)

        self.hyperparameters = hyperparameters
        self.best_overall_valid_loss = 1e10
        self.num_iteration = num_iteration
        self.best_forecast_denoise_model = nn.Module()
        self.model_path = "models_{}_{}".format(exp_name, pred_len)
        if not os.path.exists(self.model_path):
            os.makedirs(self.model_path)

        self.model_name = "{}_{}_{}_{}{}{}{}{}{}".format(forecasting_model_name, exp_name, pred_len, seed,
                                                         "_denoise",
                                                         "_gp" if gp else "",
                                                         "_predictions" if no_noise else "",
                                                         "_iso" if iso else "",
                                                         "_add_noise_only_at_training" if
                                                         add_noise_only_at_training else "")
        self.pred_len = pred_len

    def run_optuna(self):
        """
        Run the Optuna hyperparameter optimization process for the ForecastDenoising model.
        """
        study = optuna.create_study(study_name="train forecast denoise model",
                                    direction="minimize",
                                    sampler=optuna.samplers.TPESampler())
        study.optimize(self.objective, n_trials=self.n_trials,
                       n_jobs=self.n_jobs)

        pruned_trials = study.get_trials(deepcopy=False, states=[TrialState.PRUNED])
        complete_trials = study.get_trials(deepcopy=False, states=[TrialState.COMPLETE])

        print("Study statistics: ")
        print("  Number of finished trials: ", len(study.trials))
        print("  Number of pruned trials: ", len(pruned_trials))
        print("  Number of complete trials: ", len(complete_trials))

        print("Best trial:")
        trial = study.best_trial

        print("  Value: ", trial.value)

        print("  Params: ")
        for key, value in trial.params.items():
            print("    {}: {}".format(key, value))

    def objective(self, trial):
        """
        Objective function for Optuna optimization.

        Args:
        - trial: Current Optuna trial.

        Returns:
        - best_trial_valid_loss: Best validation loss achieved during the training.
        """
        param_dict = dict()
        for param, values in self.hyperparameters.items():
            param_dict[param] = trial.suggest_categorical(param, values)

        forecast_denoise_model_attributes = [attr for attr in dir(self.forecast_denoising_model) if
                                             not callable(getattr(self.forecast_denoising_model, attr))]

        for attr in forecast_denoise_model_attributes:
            if attr in self.hyperparameters.keys():
                setattr(self.forecast_denoising_model, attr, param_dict[attr])

        optimizer = Adam(self.forecast_denoising_model.parameters(), lr=param_dict["lr"])
        scheduler = torch.optim.lr_scheduler.CosineAnnealingLR(optimizer, self.num_iteration)

        best_trial_valid_loss = 1e10

        for epoch in range(self.num_epochs):
            train_loss = 0
            self.forecast_denoising_model.train()

            for train_enc, train_dec, train_y in self.train_data:
                output_fore_den, loss = self.forecast_denoising_model(train_enc, train_dec, train_y)
                train_loss += loss.item()
                optimizer.zero_grad()
                loss.backward()
                optimizer.step()
                scheduler.step()

            trial.report(loss, epoch)
            if trial.should_prune():
                raise optuna.TrialPruned()

            self.forecast_denoising_model.eval()
            valid_loss = 0

            for valid_enc, valid_dec, valid_y in self.valid_data:
                output, loss = self.forecast_denoising_model(valid_enc, valid_dec, valid_y)
                valid_loss += loss.item()

                if valid_loss < best_trial_valid_loss:
                    best_trial_valid_loss = valid_loss
                    if best_trial_valid_loss < self.best_overall_valid_loss:
                        self.best_overall_valid_loss = best_trial_valid_loss
                        self.best_forecast_denoise_model = self.forecast_denoising_model
                        torch.save({'model_state_dict': self.best_forecast_denoise_model.state_dict()},
                                   os.path.join(self.model_path, "{}".format(self.model_name)))

            return best_trial_valid_loss

    def train(self):
        """
        Train the ForecastDenoising model using Optuna hyperparameter optimization.
        """
        self.run_optuna()

    def evaluate(self):
        """
        Evaluate the performance of the best ForecastDenoising model on the test set.
        """
        self.best_forecast_denoise_model.eval()

        _, _, test_y = next(iter(self.test_data))
        total_b = len(list(iter(self.test_data)))

        predictions = torch.zeros(total_b, test_y.shape[0], self.pred_len)
        test_y_tot = torch.zeros(total_b, test_y.shape[0], self.pred_len)

        j = 0

        for test_enc, test_dec, test_y in self.test_data:
            output, _ = self.best_forecast_denoise_model(test_enc, test_dec)
            predictions[j] = output.squeeze(-1).cpu().detach()
            test_y_tot[j] = test_y[:, -self.pred_len:, :].squeeze(-1).cpu().detach()
            j += 1

        predictions = predictions.reshape(-1, 1)
        test_y = test_y_tot.reshape(-1, 1)
        normaliser = test_y.abs().mean()

        test_loss = F.mse_loss(predictions, test_y).item() / normaliser
        mse_loss = test_loss

        mae_loss = F.l1_loss(predictions, test_y).item() / normaliser
        mae_loss = mae_loss

        errors = {self.model_name: {'MSE': f"{mse_loss:.3f}", 'MAE': f"{mae_loss: .3f}"}}
        print(errors)

        error_path = "reported_errors_{}.csv".format(self.exp_name)

        df = pd.DataFrame.from_dict(errors, orient='index')

        if os.path.exists(error_path):
            df_old = pd.read_csv(error_path)
            df_new = pd.concat([df_old, df], axis=0)
            df_new.to_csv(error_path)
        else:
            df.to_csv(error_path)



import torch.nn as nn
import torch.nn.functional as F


class PoswiseFeedForwardNet(nn.Module):

    def __init__(self, d_model, d_ff):
        super(PoswiseFeedForwardNet, self).__init__()

        self.w_1 = nn.Linear(d_model, d_ff)
        self.w_2 = nn.Linear(d_ff, d_model)

    def forward(self, inputs):

        return self.w_2(F.relu(self.w_1(inputs)))


import random
import numpy as np
import pandas as pd
import torch
from torch.utils.data import TensorDataset, DataLoader


class CustomDataLoader:
    def __init__(self,
                 max_encoder_length,
                 pred_len,
                 max_train_sample,
                 max_test_sample,
                 batch_size,
                 device,
                 data,
                 target_col,
                 real_inputs):

        self.max_encoder_length = max_encoder_length
        self.pred_len = pred_len
        self.max_train_sample = max_train_sample * batch_size
        self.max_test_sample = max_test_sample * batch_size
        self.batch_size = batch_size

        seed = 1234

        torch.manual_seed(seed)
        random.seed(seed)
        np.random.seed(seed)

        total_batches = int(len(data) / self.batch_size)
        train_len = int(total_batches * batch_size * 0.6)
        valid_len = int(total_batches * batch_size * 0.2)
        test_len = int(total_batches * batch_size * 0.2)

        self.real_inputs = real_inputs

        self.num_target = 1
        self.num_features = len(real_inputs) + self.num_target - 1
        self.device = device

        real_inputs.remove(target_col)

        col_names = [target_col, "id"]
        for real in real_inputs:
            col_names.append(real)

        gen_col_names = col_names[1:]
        gen_col_names.append("value")
        gen_col_names.reverse()

        data = pd.DataFrame(data[col_names].values, columns=gen_col_names)

        train_data = data[:train_len]
        valid_data = data[train_len:train_len + valid_len]
        test_data = data[train_len + valid_len:train_len + valid_len + test_len]

        self.total_time_steps = self.max_encoder_length + self.pred_len

        self.train_loader = self.create_dataloader(train_data, max_train_sample)
        self.valid_loader = self.create_dataloader(valid_data, max_test_sample)
        self.test_loader = self.create_dataloader(test_data, max_test_sample)

        train_enc, train_dec, train_y = next(iter(self.train_loader))
        self.input_size = train_enc.shape[2]
        self.output_size = train_y.shape[2]

    def create_dataloader(self, data, max_samples):

        valid_sampling_locations, split_data_map = zip(
            *[
                (
                    (identifier, self.total_time_steps + i),
                    (identifier, df)
                )
                for identifier, df in data.groupby("id")
                if (num_entries := len(df)) >= self.total_time_steps
                for i in range(num_entries - self.total_time_steps + 1)
            ]
        )
        valid_sampling_locations = list(valid_sampling_locations)
        split_data_map = dict(split_data_map)

        ranges = [
            valid_sampling_locations[i] for i in np.random.choice(
                len(valid_sampling_locations), max_samples, replace=False)
        ]
        X = torch.zeros(max_samples, self.max_encoder_length, self.num_features)
        Y = torch.zeros(max_samples, self.pred_len, 1)

        for i, tup in enumerate(ranges):

            identifier, start_idx = tup
            sliced = split_data_map[identifier].iloc[start_idx - self.total_time_steps: start_idx]
            val = np.array(sliced.loc[:, ~sliced.columns.isin(['id'])].values, dtype=np.float64)
            val = torch.tensor(val)
            X[i] = val[:self.max_encoder_length, :]
            Y[i] = val[-self.pred_len:, 0:1]

        dataset = TensorDataset(X[:, :-self.pred_len, :], X[:, -self.pred_len:, :], Y)
        dataloader = DataLoader(dataset, batch_size=self.batch_size, shuffle=True)

        return dataloader



import torch.nn as nn
import torch
import random
import numpy as np
from modules.encoder import Encoder
from modules.decoder import Decoder


class Transformer(nn.Module):

    def __init__(self, d_model,
                 d_ff, d_k, n_heads, n_layers, device, attn_type, seed):
        super(Transformer, self).__init__()

        torch.manual_seed(seed)
        random.seed(seed)
        np.random.seed(seed)

        self.attn_type = attn_type

        self.encoder = Encoder(
            d_model=d_model, d_ff=d_ff,
            d_k=d_k, d_v=d_k, n_heads=n_heads,
            n_layers=n_layers,
            device=device, attn_type=attn_type, seed=seed)
        self.decoder = Decoder(
            d_model=d_model, d_ff=d_ff,
            d_k=d_k, d_v=d_k, n_heads=n_heads,
            n_layers=n_layers,
            device=device,
            attn_type=attn_type, seed=seed)

    def forward(self, enc_inputs, dec_inputs):

        enc_outputs = self.encoder(enc_inputs)
        dec_outputs = self.decoder(dec_inputs, enc_outputs)

        return enc_outputs, dec_outputs


import random
import numpy as np
import torch
import torch.nn as nn
from modules.multi_head_attention import MultiHeadAttention
from modules.feedforward import PoswiseFeedForwardNet
from modules.encoding import PositionalEncoding


class DecoderLayer(nn.Module):

    def __init__(self, d_model, d_ff, d_k, d_v,
                 n_heads, attn_type, seed):

        np.random.seed(seed)
        random.seed(seed)
        torch.manual_seed(seed)

        super(DecoderLayer, self).__init__()
        self.dec_self_attn = MultiHeadAttention(
            d_model=d_model, d_k=d_k,
            d_v=d_v, n_heads=n_heads,
            attn_type=attn_type, seed=seed)
        self.dec_enc_attn = MultiHeadAttention(
            d_model=d_model, d_k=d_k,
            d_v=d_v, n_heads=n_heads,
            attn_type=attn_type, seed=seed)
        self.pos_ffn = PoswiseFeedForwardNet(
            d_model=d_model, d_ff=d_ff)
        self.layer_norm = nn.LayerNorm(d_model, elementwise_affine=False)

    def forward(self, dec_inputs, enc_outputs):

        out = self.dec_self_attn(dec_inputs, dec_inputs, dec_inputs)
        out = self.layer_norm(dec_inputs + out)
        out2 = self.dec_enc_attn(out, enc_outputs, enc_outputs)
        out2 = self.layer_norm(out + out2)
        out3 = self.pos_ffn(out2)
        out3 = self.layer_norm(out2 + out3)
        return out3


class Decoder(nn.Module):

    def __init__(self, d_model, d_ff, d_k, d_v,
                 n_heads, n_layers, device,
                 attn_type, seed):
        super(Decoder, self).__init__()

        np.random.seed(seed)
        random.seed(seed)
        torch.manual_seed(seed)

        self.device = device
        self.attn_type = attn_type
        self.pos_emb = PositionalEncoding(
            d_hid=d_model, device=device)
        self.layer_norm = nn.LayerNorm(d_model)
        self.layers = []
        for _ in range(n_layers):
            decoder_layer = DecoderLayer(
                d_model=d_model, d_ff=d_ff,
                d_k=d_k, d_v=d_v,
                n_heads=n_heads,
                attn_type=attn_type, seed=seed)
            self.layers.append(decoder_layer)
        self.layers = nn.ModuleList(self.layers)
        self.d_k = d_k

    def forward(self, dec_inputs, enc_outputs):

        dec_outputs = self.pos_emb(dec_inputs)

        for layer in self.layers:
            dec_outputs = layer(
                dec_inputs=dec_outputs,
                enc_outputs=enc_outputs
            )

        return dec_outputs


import argparse
import pandas as pd
import torch
import dataformatter
from modules.transformer import Transformer
from data_loader import CustomDataLoader
from train_forecast_blur_denoise import TrainForecastBlurDenoise
from torch.utils.data import DataLoader, TensorDataset


class TimeSeriesDataset:
    def __init__(self, num_samples,
                 input_sequence_length,
                 output_sequence_length,
                 input_size,
                 output_size):

        self.num_samples = num_samples
        self.input_sequence_length = input_sequence_length
        self.output_sequence_length = output_sequence_length
        self.input_size = input_size
        self.output_size = output_size

        # input for encoder, input for decoder, and output (ground-truth)

        self.data = TensorDataset(torch.randn(self.num_samples, self.input_sequence_length, self.input_size),
                                  torch.randn(self.num_samples, self.output_sequence_length, self.input_size),
                                  torch.randn(self.num_samples, self.output_sequence_length, self.output_size))


def get_dataset(args, device):

    # Target column names for different datasets

    target_col = {"traffic": "values",
                  "electricity": "power_usage",
                  "exchange": "OT",
                  "solar": "Power(MW)",
                  "air_quality": "NO2",
                  "watershed": "Conductivity"}

    data_formatter = dataformatter.DataFormatter(args.exp_name)

    data_path = args.data_path
    df = pd.read_csv(data_path)
    df.sort_values(by=["id", "hours_from_start"], inplace=True)
    data = data_formatter.transform_data(df)

    # Data loader configuration (replace with your own dataloader)

    data_loader = CustomDataLoader(max_encoder_length=args.max_encoder_length,
                                   pred_len=args.pred_len,
                                   data=data,
                                   real_inputs=data_formatter.real_inputs,
                                   target_col=target_col[args.exp_name],
                                   max_train_sample=args.max_train_sample,
                                   max_test_sample=args.max_test_sample,
                                   batch_size=args.batch_size,
                                   device=device)

    return data_loader


def main():
    """
    An example to execute the training and evaluation workflow for the forecasting and denoising model.
    """
    # Argument parser for configuring the training process

    parser = argparse.ArgumentParser(description="forecast-denoise argument parser")

    # Forecasting model parameters

    parser.add_argument("--forecasting_model_name", type=str, default="autoformer")
    parser.add_argument("--exp_name", type=str, default='solar')
    parser.add_argument("--cuda", type=str, default="cuda:0")
    parser.add_argument("--noise_type", type=str, default="gp")
    parser.add_argument("--seed", type=int, default=1234)
    parser.add_argument("--pred_len", type=int, default=96)
    parser.add_argument("--max_encoder_length", type=int, default=192)
    parser.add_argument("--max_train_sample", type=int, default=32000)
    parser.add_argument("--max_test_sample", type=int, default=3840)
    parser.add_argument("--batch_size", type=int, default=256)
    parser.add_argument("--n_trials", type=int, default=50)
    parser.add_argument("--n_jobs", type=int, default=1)
    parser.add_argument("--num_inducing", type=int, default=1)
    parser.add_argument("--learning_residual", type=lambda x: str(x).lower() == "true", default="False")
    parser.add_argument("--no-noise", type=lambda x: str(x).lower() == "true", default="False")
    parser.add_argument("--add_noise_only_at_training", type=lambda x: str(x).lower() == "true", default="False")
    parser.add_argument("--num_epochs", type=int, default=10)
    parser.add_argument("--data_path", type=str, default="~/research/Corruption-resilient-Forecasting-Models/solar.csv")

    args = parser.parse_args()

    # Device configuration
    device = torch.device(args.cuda if torch.cuda.is_available() else "cpu")

    if args.exp_name == "toy_data":
        # example parameters
        num_samples_train = 32
        num_samples_valid = 8
        num_samples_test = 8
        input_sequence_length = 96
        output_sequence_length = 96
        batch_size = 4
        input_size = 5
        output_size = 1

        train_dataset = TimeSeriesDataset(num_samples_train, input_sequence_length,
                                          output_sequence_length, input_size, output_size)
        valid_dataset = TimeSeriesDataset(num_samples_valid, input_sequence_length,
                                          output_sequence_length, input_size, output_size)
        test_dataset = TimeSeriesDataset(num_samples_test, input_sequence_length,
                                         output_sequence_length, input_size, output_size)

        train_loader = DataLoader(train_dataset.data, batch_size=batch_size, shuffle=True)
        valid_loader = DataLoader(valid_dataset.data, batch_size=batch_size, shuffle=False)
        test_loader = DataLoader(test_dataset.data, batch_size=batch_size, shuffle=False)

    else:

        data_loader = get_dataset(args, device)
        train_loader = data_loader.train_loader
        valid_loader = data_loader.valid_loader
        test_loader = data_loader.test_loader
        input_size = data_loader.input_size
        output_size = data_loader.output_size

    # Dimensionality of the model

    d_model = 32

    # Initializing the forecasting model (replace with your forecasting model)
    forecasting_model = Transformer(d_model=d_model,
                                    d_ff=d_model * 4,
                                    d_k=d_model, n_heads=8,
                                    n_layers=1, device=device,
                                    attn_type="autoformer",
                                    seed=1234).to(device)

    # Hyperparameter search space (change accordingly)

    hyperparameters = {"d_model": [16, 32], "n_heads": [1, 8],
                       "n_layers": [1, 2], "lr": [0.01, 0.001],
                       "num_inducing": [32, 64]}

    # Initializing and training the ForecastDenoise model using Optuna

    trainforecastdenoise = TrainForecastBlurDenoise(forecasting_model=forecasting_model,
                                                    train=train_loader,
                                                    valid=valid_loader,
                                                    test=test_loader,
                                                    num_iteration=args.max_encoder_length,
                                                    noise_type="gp",
                                                    num_inducing=32,
                                                    add_noise_only_at_training=args.add_noise_only_at_training,
                                                    input_size=input_size,
                                                    output_size=output_size,
                                                    pred_len=96,
                                                    hyperparameters=hyperparameters,
                                                    seed=1234,
                                                    device=device)

    trainforecastdenoise.train()
    trainforecastdenoise.evaluate()


if __name__ == '__main__':
    main()



from setuptools import setup, find_packages
from forecastblurdenoise import __version__
setup(
    name='forecastblurdenoise',
    version=__version__,
    author='Sepideh Koohfar',
    packages=find_packages(),
    python_requires='>=3.9',
    install_requires=[
        'gpytorch>=1.9.0',
        'torch>=2.0.1',
        'pandas>=1.5.2',
        'numpy>=1.23.5',
        'optuna>=3.3.0'
    ],
    license='MIT',
    author_email='sepideh.koohfar@unh.edu',
    url='https://github.com/SepKfr/Fine_grained_forecasting',
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown',
    classifiers=[
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.9',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        "Topic :: Scientific/Engineering :: Artificial Intelligence"],
    entry_points={
        'console_scripts': [
             'example_usage=forecastblurdenoise.main:main'
         ],
    }
)



import torch
import torch.nn as nn
from gpytorch.mlls import DeepApproximateMLL, VariationalELBO
from DeepGP import DeepGPp


class BlurDenoiseModel(nn.Module):
    def __init__(self, model, d_model, num_inducing, gp, no_noise=False, iso=False):
        """
        Blur and Denoise model.

        Args:
        - model (nn.Module): Underlying forecasting model for adding and removing noise.
        - d_model (int): Dimensionality of the model.
        - num_inducing (int): Number of inducing points for the GP model.
        - gp (bool): Flag indicating whether to use GP as the blur model.
        - no_noise (bool): Flag indicating whether to add no noise during
          denoising (denoise predictions directly).
        - iso (bool): Flag indicating whether to use isotropic noise.
        """
        super(BlurDenoiseModel, self).__init__()

        self.denoising_model = model

        # Initialize DeepGP model for GP regression
        self.deep_gp = DeepGPp(d_model, num_inducing)
        self.gp = gp
        self.sigma = nn.Parameter(torch.randn(1))

        # Layer normalization and feedforward networks
        self.norm_1 = nn.LayerNorm(d_model)
        self.norm_2 = nn.LayerNorm(d_model)
        self.ffn_1 = nn.Sequential(nn.Linear(d_model, d_model*4),
                                   nn.ReLU(),
                                   nn.Linear(d_model*4, d_model))
        self.ffn_2 = nn.Sequential(nn.Linear(d_model, d_model*4),
                                   nn.ReLU(),
                                   nn.Linear(d_model*4, d_model))

        self.d = d_model
        self.no_noise = no_noise
        self.iso = iso

    def add_gp_noise(self, x):
        """
        Add GP noise to the input using the DeepGP model.

        Args:
        - x (Tensor): Input tensor.

        Returns:
        - x_noisy (Tensor): Noisy input with added GP noise.
        - dist (Tensor): GP distribution if GP is used.
        """
        b, s, _ = x.shape

        # Predict GP noise and apply layer normalization
        eps_gp, dist = self.deep_gp.predict(x)
        x_noisy = self.norm_1(x + self.ffn_1(eps_gp))

        return x_noisy, dist

    def forward(self, enc_inputs, dec_inputs):
        """
        Forward pass of the BlurDenoiseModel.

        Args:
        - enc_inputs (Tensor): Encoder inputs.
        - dec_inputs (Tensor): Decoder inputs.

        Returns:
        - dec_output (Tensor): Denoised decoder output.
        - dist (Tensor): GP distribution if GP is used.
        """
        eps_enc = torch.randn_like(enc_inputs)
        eps_dec = torch.randn_like(dec_inputs)
        dist = None

        if self.gp:

            # Add GP noise to encoder and decoder inputs
            enc_noisy, dist_enc = self.add_gp_noise(enc_inputs)
            dec_noisy, dist = self.add_gp_noise(dec_inputs)

        elif self.iso:
            # Add scaled isotropic noise with trainable scale
            enc_noisy = enc_inputs.add_(eps_enc * torch.clip(self.sigma, 0, 0.1))
            dec_noisy = dec_inputs.add_(eps_dec * torch.clip(self.sigma, 0, 0.1))
        else:
            # No noise addition
            enc_noisy = enc_inputs
            dec_noisy = dec_inputs

        # Perform denoising with the underlying forecasting model
        enc_denoise, dec_denoise = self.denoising_model(enc_noisy, dec_noisy)

        # Apply layer normalization and feedforward network to the decoder output
        dec_output = self.norm_2(dec_inputs + self.ffn_2(dec_denoise))

        return dec_output, dist


class ForecastBlurDenoise(nn.Module):
    def __init__(self, forecasting_model: nn.Module,
                 gp: bool,
                 iso: bool,
                 no_noise: bool,
                 add_noise_only_at_training: bool,
                 pred_len: int,
                 input_size: int,
                 output_size: int,
                 num_inducing: int,
                 d_model: int = 32):
        """
        Forecast-blur-denoise Module.

        Args:
        - forecasting_model (nn.Module): The underlying forecasting model.
        - gp (bool): Flag indicating whether to use GP as the blur model.
        - iso (bool): Flag indicating whether to use Gaussian isotropic for the blur model.
        - no_noise (bool): Flag indicating whether to add no noise during denoising
         (denoise predictions directly).
        - add_noise_only_at_training (bool): Flag indicating whether to add noise only during training.
        - pred_len (int): Length of the prediction horizon.
        - src_input_size (int): Number of features in input.
        - tgt_output_size (int): Number of features in output.
        - num_inducing (int): Number of inducing points for GP model.
        - d_model (int): Dimensionality of the model (default is 32).
        """
        super(ForecastBlurDenoise, self).__init__()

        self.pred_len = pred_len
        self.add_noise_only_at_training = add_noise_only_at_training
        self.gp = gp
        self.lam = nn.Parameter(torch.randn(1))

        self.forecasting_model = forecasting_model

        # Initialize the blur and denoise model
        self.de_model = BlurDenoiseModel(self.forecasting_model,
                                         d_model,
                                         gp=gp,
                                         no_noise=no_noise,
                                         iso=iso,
                                         num_inducing=num_inducing)

        self.final_projection = nn.Linear(d_model, output_size)
        self.enc_embedding = nn.Linear(input_size, d_model)
        self.dec_embedding = nn.Linear(input_size, d_model)
        self.d_model = d_model

    def forward(self, enc_inputs, dec_inputs, y_true=None):
        """
        Forward pass of the ForecastDenoising model.

        Args:
        - enc_inputs (Tensor): Encoder inputs.
        - dec_inputs (Tensor): Decoder inputs.
        - y_true (Tensor): True labels for training (optional).

        Returns:
        - final_outputs (Tensor): Model's final predictions.
        - loss (Tensor): Combined loss from denoising and forecasting components.
        """
        mll_error = 0
        loss = 0

        # Indicate whether to perform denoising
        denoise = True

        '''
        If add_noise_only_at_training flag is set and during test 
        do not perform denoising
        '''

        if self.add_noise_only_at_training and not self.training:
            denoise = False

        # Embed inputs
        enc_inputs = self.enc_embedding(enc_inputs)
        dec_inputs = self.dec_embedding(dec_inputs)

        # Get outputs from the forecasting model
        enc_outputs, dec_outputs = self.forecasting_model(enc_inputs, dec_inputs)

        if denoise:
            # Apply denoising
            de_model_outputs, dist = self.de_model(enc_outputs.clone(), dec_outputs.clone())
        else:
            de_model_outputs = dec_outputs

        # Project the denoised outputs to the final output space
        final_outputs = self.final_projection(de_model_outputs[:, -self.pred_len:, :])

        # If using GP and during training, compute MLL loss
        if self.gp and self.training:

            mll = DeepApproximateMLL(
                VariationalELBO(self.de_model.deep_gp.likelihood, self.de_model.deep_gp, self.d_model))

            mll_error = -mll(dist, y_true).mean()

        # If ground truth is available, compute MSE loss and combine with MLL loss
        if y_true is not None:
            mse_loss = nn.MSELoss()(y_true, final_outputs)
            loss = mse_loss + self.lam * mll_error

        return final_outputs, loss



# coding=utf-8
# Copyright 2021 The Google Research Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# Lint as: python3
import pandas as pd
import sklearn.preprocessing


class DataAttributes:

    @property
    def _data_def(self):

        return {"traffic": {'id': 'id', 'target': 'values', 'covariates': []},
                "electricity": {'id': 'id', 'target': 'power_usage', 'covariates': []},
                "solar": {'id': 'id', 'target': 'Power(MW)',
                 'covariates': [], 'time': 'hours_from_start'},
                "air_quality": {'id': 'id', 'target': 'NO2',
                       'covariates': ['CO', 'TEMP']},
                "watershed": {'id': 'id', 'target': 'Conductivity',
                     'covariates': ['Q']},
                "exchange": {'id': 'id', 'target': 'OT',
                    'covariates': ['0', '1', '2', '3', '4', '5']}}

    def __init__(self, exp_name):

        col_def = self._data_def[exp_name]
        self.id = col_def['id']
        self.target = col_def['target']
        self.covariates = col_def['covariates']


class DataFormatter:
    """Defines and formats data_set for the electricity dataset.
    Note that per-entity z-score normalization is used here, and is implemented
    across functions.
    Attributes:
    column_definition: Defines input and data_set type of column used in the
      experiment.
    identifiers: Entity identifiers used in experiments.
    """

    def __init__(self, exp_name: str):
        """Initialises formatter."""

        self.identifiers = None
        self.real_scalers = None
        self.target_scaler = None

        self.column_definition = DataAttributes(exp_name)
        self.id_column = self.column_definition.id
        self.target_column = self.column_definition.target

        self.real_inputs = []
        self.real_inputs.append(self.target_column)
        for covar in self.column_definition.covariates:
            self.real_inputs.append(covar)

        print(self.real_inputs)

    def transform_data(self, df):

        print('Formatting data.')

        self.set_scalers(df)

        return self.transform_inputs(df)

    def set_scalers(self, df):

        """Calibrates scalers using the data_set supplied.
        Args:
          df: Data to use to calibrate scalers.
        """
        print('Setting scalers with data_set...')

        self.real_scalers = {}
        self.target_scaler = {}
        identifiers = []

        for identifier, sliced in df.groupby(self.id_column):

            data = sliced[self.real_inputs].values
            targets = sliced[self.target_column].values
            if len(data.shape) == 1:
                data = data.reshape(-1, 1)
            if len(targets.shape) == 1:
                targets = targets.reshape(-1, 1)

            self.real_scalers[identifier] = sklearn.preprocessing.StandardScaler().fit(data)

            self.target_scaler[identifier] = sklearn.preprocessing.StandardScaler().fit(targets)

            identifiers.append(identifier)

        # Extract identifiers in case required
        self.identifiers = identifiers

    def transform_inputs(self, df):
        """Performs feature transformations.
        This includes both feature engineering, preprocessing and normalisation.
        Args:
          df: Data frame to transform.
        Returns:
          Transformed data_set frame.
        """

        if self.real_scalers is None:
            raise ValueError('Scalers have not been set!')

        # Transform real inputs per entity
        df_list = []
        for identifier, sliced in df.groupby(self.id_column):
            # Filter out any trajectories that are too short
            sliced_copy = sliced.copy()
            sliced_copy[self.real_inputs] = self.real_scalers[identifier].transform(
                sliced_copy[self.real_inputs].values)
            df_list.append(sliced_copy)

        output = pd.concat(df_list, axis=0)

        return output

    def format_predictions(self, predictions):
        """Reverts any normalisation to give predictions in original scale.
        Args:
          predictions: Dataframe of model predictions.
        Returns:
          Data frame of unnormalised predictions.
        """

        if self.target_scaler is None:
            raise ValueError('Scalers have not been set!')

        column_names = predictions.columns

        df_list = []
        for identifier, sliced in predictions.groupby('identifier'):
            sliced_copy = sliced.copy()
            target_scaler = self.target_scaler[identifier]

            for col in column_names:
                if col not in {'identifier'}:
                    try:
                        sliced_copy[col] = target_scaler.inverse_transform(sliced_copy[col])
                    except ValueError:
                        if len(sliced_copy[col]) == 1:
                            pred = sliced_copy[col].to_numpy().reshape(1, -1)
                        else:
                            pred = sliced_copy[col].to_numpy().reshape(-1, 1)

                        sliced_copy[col] = target_scaler.inverse_transform(pred)

            df_list.append(sliced_copy)
        if len(df_list) == 0:
            output = None
        else:
            output = pd.concat(df_list, axis=0)

        return output


import torch
from gpytorch.distributions import MultivariateNormal, MultitaskMultivariateNormal
from gpytorch.kernels import ScaleKernel, RBFKernel
from gpytorch.likelihoods import MultitaskGaussianLikelihood
from gpytorch.means import ConstantMean, LinearMean
from gpytorch.models.deep_gps import DeepGPLayer, DeepGP
from gpytorch.variational import VariationalStrategy, MeanFieldVariationalDistribution


class ToyDeepGPHiddenLayer(DeepGPLayer):
    def __init__(self, input_dims, output_dims, num_inducing=32, mean_type='constant'):

        if output_dims is None:
            inducing_points = torch.randn(num_inducing, input_dims)
            batch_shape = torch.Size([])
        else:
            inducing_points = torch.randn(output_dims, num_inducing, input_dims)
            batch_shape = torch.Size([output_dims])

        variational_distribution = MeanFieldVariationalDistribution(
            num_inducing_points=num_inducing,
            batch_shape=batch_shape
        )

        variational_strategy = VariationalStrategy(
            self,
            inducing_points,
            variational_distribution,
            learn_inducing_locations=True
        )

        super(ToyDeepGPHiddenLayer, self).__init__(variational_strategy, input_dims, output_dims)

        if mean_type == 'constant':
            self.mean_module = ConstantMean(batch_shape=batch_shape)
        else:
            self.mean_module = LinearMean(input_dims)
        self.covar_module = ScaleKernel(
            RBFKernel(batch_shape=batch_shape, ard_num_dims=input_dims),
            batch_shape=batch_shape, ard_num_dims=None
        )

    def forward(self, x):
        mean_x = self.mean_module(x)
        covar_x = self.covar_module(x)
        return MultivariateNormal(mean_x, covar_x)


class DeepGPp(DeepGP):
    def __init__(self, num_hidden_dims, num_inducing):

        hidden_layer = ToyDeepGPHiddenLayer(
            input_dims=num_hidden_dims,
            output_dims=num_hidden_dims,
            mean_type='linear',
            num_inducing=num_inducing
        )

        super().__init__()

        self.hidden_layer = hidden_layer
        self.likelihood = MultitaskGaussianLikelihood(num_tasks=num_hidden_dims)

    def forward(self, inputs):

        dist = self.hidden_layer(inputs)
        return dist

    def predict(self, x):

        dist = self(x)
        preds = self.likelihood(dist)
        preds_mean = preds.mean.mean(0)

        return preds_mean, dist